/*
 * Buxton
 *
 * Copyright (C) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <assert.h>

#include <glib.h>

#include "log.h"
#include "config.h"

#define K_TYPE     "Type"
#define K_BACKEND  "Backend"
#define K_STORAGE  "Storage"
#define K_DESC     "Description"

#define K_T_SYSTEM "System"
#define K_T_USER   "User"

#define K_B_DB     "persistent"
#define K_B_MEM    "volatile"

#define G_LOG_CONF "log-config"
#define K_LOG_ON   "OnOff"
#define K_LOG_MAX  "MaxLine"

static GHashTable *layers;

static void free_layer(struct layer *layer)
{
	if (!layer)
		return;

	g_free(layer->name);
	g_free(layer->backend);
	g_free(layer->description);
	free(layer);
}

static int load_conf(const char *confpath, GKeyFile **keyfile)
{
	GKeyFile *kf;
	gboolean b;
	GError *err;

	if (!confpath || !keyfile) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	kf = g_key_file_new();
	if (!kf) {
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	err = NULL;
	b = g_key_file_load_from_file(kf, confpath, G_KEY_FILE_NONE, &err);
	if (!b) {
		bxt_err("Load '%s' error: %s", confpath,
				err ? err->message : "");
		g_clear_error(&err);
		g_key_file_free(kf);
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	*keyfile = kf;

	return BUXTON_ERROR_NONE;
}

static int create_layer(GKeyFile *kf, gchar *name, struct layer **new_layer)
{
	GError *err;
	struct layer *layer;
	gchar *s;
	int ret;

	if (!name || !*name || !new_layer) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	layer = calloc(1, sizeof(*layer));
	if (!layer) {
		bxt_err("out of memory");
		ret = BUXTON_ERROR_OUT_OF_MEMORY;
		goto err;
	}

	/* 'Type' */
	err = NULL;
	s = g_key_file_get_string(kf, name, K_TYPE, &err);
	if (!s) {
		bxt_err("Layer '%s' : %s",
				name, err ? err->message : "");
		g_clear_error(&err);
		ret = BUXTON_ERROR_INVALID_PARAMETER;
		goto err;
	}

	if (!strncmp(s, K_T_SYSTEM, sizeof(K_T_SYSTEM)))
		layer->type = LAYER_SYSTEM;
	else if (!strncmp(s, K_T_USER, sizeof(K_T_USER)))
		layer->type = LAYER_USER;

	g_free(s);

	/* 'Backend' */
	s = g_key_file_get_string(kf, name, K_BACKEND, &err);
	if (!s) {
		bxt_err("Layer '%s' : %s",
				name, err ? err->message : "");
		g_clear_error(&err);
		ret = BUXTON_ERROR_INVALID_PARAMETER;
		goto err;
	}

	layer->backend = s;

	/* 'Storage' */
	s = g_key_file_get_string(kf, name, K_STORAGE, &err);
	if (!s) {
		bxt_err("Layer '%s' : %s",
				name, err ? err->message : "");
		g_clear_error(&err);
		ret = BUXTON_ERROR_INVALID_PARAMETER;
		goto err;
	}

	if (!strncasecmp(s, K_B_DB, sizeof(K_B_DB)))
		layer->storage = STORAGE_PERSISTENT;
	else if (!strncasecmp(s, K_B_MEM, sizeof(K_B_MEM)))
		layer->storage = STORAGE_VOLATILE;

	g_free(s);

	/* 'Description' */
	s = g_key_file_get_string(kf, name, K_DESC, &err);
	if (!s) {
		bxt_err("Layer '%s' : %s",
				name, err ? err->message : "");
		g_clear_error(&err);
		ret = BUXTON_ERROR_INVALID_PARAMETER;
		goto err;
	}

	layer->description = s;

	/* Layer name */
	layer->name = name;
	*new_layer = layer;

	return BUXTON_ERROR_NONE;

err:
	g_free(name);
	free_layer(layer);

	return ret;
}

static gboolean has_backend(GList *backends, const char *name)
{
	GList *l;

	if (!name || !*name)
		return FALSE;

	for (l = backends; l; l = g_list_next(l)) {
		const char *nm = l->data;

		if (nm && !strcmp(name, nm))
			return TRUE;
	}

	return FALSE;
}

static void add_layers(GKeyFile *kf, GList *backends)
{
	gchar **lays;
	int i, ret;
	struct layer *layer;
	struct layer *f;
	gboolean b;

	if (!kf || !backends) {
		bxt_err("Invalid parameter");
		return;
	}

	if (!layers) {
		bxt_err("layers is not initilized.");
		return;
	}

	lays = g_key_file_get_groups(kf, NULL);
	if (!lays) {
		bxt_err("No specified layers");
		return;
	}

	i = 0;
	while (lays[i]) {
		ret = create_layer(kf, lays[i++], &layer);
		if (ret != BUXTON_ERROR_NONE)
			continue;

		b = has_backend(backends, layer->backend);
		if (!b) {
			bxt_err("Layer '%s' : invalid backend", layer->name);
			free_layer(layer);
			continue;
		}

		f = g_hash_table_lookup(layers, layer->name);
		if (f) {
			bxt_err("Layer '%s' : already exists", layer->name);
			free_layer(layer);
			continue;
		}

		if (layer->type == LAYER_UNKNWON) {
			bxt_err("Layer '%s' : unknwon type", layer->name);
			free_layer(layer);
			continue;
		}

		if (layer->storage == STORAGE_UNKNOWN) {
			bxt_err("Layer '%s' : unknwon storage type",
					layer->name);
			free_layer(layer);
			continue;
		}

		g_hash_table_insert(layers, layer->name, layer);
	}

	g_free(lays);
}

static void load_log_conf(GKeyFile *kf, bool *log_on, int *max_line)
{
	GError *err = NULL;
	gchar *buf = NULL;
	bool onoff;
	int max = 0;

	if (!kf || !log_on || !max_line) {
		bxt_err("Invalid parameter");
		return;
	}

	/* 'OnOff' */
	buf = g_key_file_get_string(kf, G_LOG_CONF, K_LOG_ON, &err);
	if (!buf) {
		bxt_err("get_string failed :Group '%s', Key '%s' : %s",
				G_LOG_CONF, K_LOG_ON, err ? err->message : "");
		goto out;
	}
	if (strcasecmp(buf, "ON") == 0) {
		onoff = true;
	} else if (strcasecmp(buf, "OFF") == 0) {
		onoff = false;
	} else {
		bxt_err("get_string failed : Group '%s', Key '%s' : '%s'(len:%zu)",
				G_LOG_CONF, K_LOG_ON, buf, strlen(buf));
		goto out;
	}

	if (onoff) {
		g_free(buf);
		buf = NULL;

		/* max line */
		buf = g_key_file_get_string(kf, G_LOG_CONF, K_LOG_MAX, &err);
		if (!buf) {
			bxt_err("get_string failed : Group '%s', Key '%s' : '%s'",
					G_LOG_CONF, K_LOG_MAX, err ? err->message : "");
			goto out;
		}

		max = atoi(buf);
		if (max <= 0) {
			bxt_err("atoi failed : Group '%s', Key '%s' : '%s'",
				G_LOG_CONF, K_LOG_MAX, buf);
			goto out;
		}
	}

	*log_on = onoff;
	*max_line = max;
	bxt_err("log config log on : %s , max line %d", onoff ? "True" : "False", max);

out:
	if (err)
		g_clear_error(&err);
	if (buf)
		g_free(buf);

	g_key_file_remove_group(kf, G_LOG_CONF, NULL);
}

int conf_get_layer(const char *name, const struct layer **layer)
{
	if (!name || !*name || !layer) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	if (!layers) {
		bxt_err("layers is not initilized.");
		return BUXTON_ERROR_INVALID_OPERATION;
	}

	*layer = g_hash_table_lookup(layers, name);
	if (!*layer) {
		bxt_dbg("Layer '%s' not exist", name);
		return BUXTON_ERROR_NOT_EXIST;
	}

	return BUXTON_ERROR_NONE;
}

int conf_remove(const char *name)
{
	const struct layer *layer;
	int ret;

	ret = conf_get_layer(name, &layer);
	if (ret != BUXTON_ERROR_NONE)
		return ret;

	g_hash_table_remove(layers, name);

	return BUXTON_ERROR_NONE;
}

void conf_exit(void)
{
	g_hash_table_destroy(layers);
	layers = NULL;
}

int conf_init(const char *confpath, GList *backends, bool *log_on, int *max_line)
{
	GKeyFile *kf;
	int ret;

	if (!confpath || !*confpath || !log_on || !max_line) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	if (layers)
		return BUXTON_ERROR_NONE;

	ret = load_conf(confpath, &kf);
	if (ret != BUXTON_ERROR_NONE)
		return ret;

	layers = g_hash_table_new_full(g_str_hash, g_str_equal,
			NULL, (GDestroyNotify)free_layer);
	if (!layers) {
		bxt_err("out of memory");
		g_key_file_free(kf);
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	load_log_conf(kf, log_on, max_line);

	add_layers(kf, backends);

	g_key_file_free(kf);

	return BUXTON_ERROR_NONE;
}
